import os

if os.name == 'nt':
    import winsound

def play_beep():
    if os.name == 'nt':
        winsound.PlaySound("SystemQuestion", winsound.SND_ASYNC)
    else:
        print('\a')


def beep_input(prompt, should_beep = True):
    if should_beep:
        play_beep()
    return input(prompt)




import os
import traceback
from joblib import Parallel, delayed

from .profile import Profile

from .schedule_utils import ScheduleUtils
from .create_driver import BrowserConfig, create_driver
from .bose_driver import BoseDriver
from .utils import relative_path,merge_dicts_in_one_dict, write_file, write_html, write_json,get_driver_path
from .local_storage import LocalStorage
from .analytics import Analytics
from .task_info import TaskInfo
from .task_config import TaskConfig

class RetryException(Exception):
    pass

def get_driver_url_safe(driver):
    try:
        return driver.current_url
    except:
        return "Failed to get driver url"

def get_page_source_safe(driver):
    try:
        return driver.page_source
    except:
        return "Failed to get page_source"


def _download_driver():
    from .download_driver import download_driver
    download_driver()

class BaseTask():
    def __init__(self):
        self.task_path = None
        self.task_id = None        


    task_config = TaskConfig()

    def get_task_config(self):
        return self.task_config

    browser_config = BrowserConfig()

    def get_browser_config(self, data):
        return self.browser_config


    data = [None]
    def get_data(self):
        return self.data


    def schedule(self, data):
        """
            Seconds delay between each run
        """
        return ScheduleUtils.no_delay(data)



        
    def set_config_on_driver(self, driver):
            driver.task_id = self.task_id
            driver.task_path = self.task_path
            driver.beep = self._task_config.beep

    def create_driver(self, config: BrowserConfig):
        driver =  create_driver(config)
        self.set_config_on_driver(driver)
        return driver
    
    # simple headless drivers no profile options
    def parallel(self, callable, data_list= [None], n = 2):
        def run(data):
            config = self.get_browser_config(data)
            driver = self.create_driver(config)
            
            result = callable(driver, data)
            
            driver.close()
            return result
        
        n = min(len(data_list), n)

        result = (Parallel(n_jobs=n, backend="threading")(delayed(run)(l) for l in data_list))
        return result 
        

    def begin_task(self, data, task_config:TaskConfig):
        def create_directories(task_path):

            driver_path =  relative_path(get_driver_path(), 0)
            if not os.path.isfile(driver_path):
                _download_driver()

            tasks_path =  relative_path('tasks/', 0)
            if not os.path.exists(tasks_path):
                os.makedirs(tasks_path)

            output_path =  relative_path('output/', 0)
            if not os.path.exists(output_path):
                os.makedirs(output_path)

            profiles_path =  relative_path('profiles/', 0)
            if not os.path.exists(profiles_path):
                os.makedirs(profiles_path)

            path =  relative_path(task_path, 0)
            if not os.path.exists(path):
                os.makedirs(path)

        
        def run_task(is_retry, retry_attempt):
            # print('Launching Task')
            task = TaskInfo()
            task_name = self.__class__.__name__
            task.set_task_name(task_name)

            final_image = "final.png"
            def end_task(driver:BoseDriver):
                task.end()
                task.set_ip()
                data = task.get_data()
                driver.save_screenshot(final_image)
                
                html_path  = f'{self.task_path}/page.html'
                source = get_page_source_safe(driver)
                write_html(source, html_path)

                data["last_url"] = get_driver_url_safe(driver)
                
                if is_retry: 
                    data["is_retry"] = is_retry
                    data["retry_attempt"] = retry_attempt


                task_info_path  = f'{self.task_path}/task_info.json'
                
                if driver._init_data is not None:
                    data = merge_dicts_in_one_dict(data , driver._init_data)
                
                write_json(data , task_info_path)
                Analytics.send_tracking_data(task_name)
            count = LocalStorage.get_item('count', 0) + 1
            LocalStorage.set_item('count', count)

            self.task_path = f'tasks/{count}' 
            self.task_id = count

            create_directories(self.task_path)
            
            task.start()

            self._task_config = task_config 
            config = self.get_browser_config(data)
            driver = self.create_driver(config)

            if config.profile is not None:
                Profile.profile = config.profile


            self.set_config_on_driver(driver)

            self.beep = task_config.beep

            final_image_path = f'{self.task_path}/{final_image}'
            
            def close_driver(driver:BoseDriver):
                print("Closing Browser")                
                # set tiny profile data
                driver.close()
                print("Closed Browser")  

                if self.task_config.log_time:
                    duration = task.get_data()['duration']
                    print(f'Task done in {duration}.')              

            result = None
            try:
                result = self.run(driver, data)
                end_task(driver)
                close_driver(driver)
                print(f'View Final Screenshot at {final_image_path}')
                return result
            except RetryException as error:
                end_task(driver)
                close_driver(driver)
                print('Task is being Retried!')
                return run_task(True, retry_attempt + 1)
            except Exception as error:
                exception_log = traceback.format_exc()
                traceback.print_exc()
                end_task(driver)
                
                error_log_path  = f'{self.task_path}/error.log'
                write_file(exception_log, error_log_path)

                IS_PRODUCTION = os.environ.get("ENV") == "production"

                if not IS_PRODUCTION:
                    if not task_config.close_on_crash:
                        driver.prompt("Press Enter To Close Browser")
                    
                close_driver(driver)

                print(f'Task Failed! View Final Screenshot at {final_image_path}')
                return result

        final = run_task(False, 0)

        Profile.profile = None

        return final

    def run(self, driver: BoseDriver, data: any):
        pass

    def is_new_user(self):
        count = LocalStorage.get_item('count', 0)
        return count  <= 5




import string
import requests
import random
from .temp_mail import TempMail
from .utils import merge_list_of_dicts, merge_dicts_in_one_dict

from datetime import datetime

def convert_timestamp(timestamp):
    dt = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%fZ")
    year = dt.year
    month = dt.month
    day = dt.day
    return {"year": year, "month": month, "day": day}


class Gender():
    MEN = 'male'
    BOTH = ''
    WOMEN = 'female'

class Country():
    AU = 'AU'
    BR = 'BR'
    CA = 'CA'
    CH = 'CH'
    DE = 'DE'
    DK = 'DK'
    ES = 'ES'
    FI = 'FI'
    FR = 'FR'
    GB = 'GB'
    IE = 'IE'
    IN = 'IN'
    IR = 'IR'
    MX = 'MX'
    NL = 'NL'
    NO = 'NO'
    NZ = 'NZ'
    RS = 'RS'
    TR = 'TR'
    UA = 'UA'
    US = 'US'
class ProblemWithGetPersonData(Exception):
    """Problem with obtaining personal data"""


def get_missing_chars(word:str):
    missing_chars = []
    
    if not any(char.isupper() for char in word):
        missing_chars.append(random.choice(string.ascii_uppercase))
    
    if not any(char.islower() for char in word):
        missing_chars.append(random.choice(string.ascii_lowercase))
    
    if not any(char.isdigit() for char in word):
        missing_chars.append(random.choice(string.digits))
    
    if not any(char in string.punctuation for char in word):
        missing_chars.append(random.choice(string.punctuation))
    
    random.shuffle(missing_chars)
    return ''.join(missing_chars)

def process_password(word):
    word = word + get_missing_chars(word)

    if len(word) < 8:
        digits_needed = 8 - len(word)
        
        # Generate a random number with the required number of digits
        random_number = random.randint(10 ** (digits_needed - 1), (10 ** digits_needed) - 1)
        
        # Convert the number to a string and append it to the word
        word = word + str(random_number)


    return word

def __get_person(user_data: dict):
    email = user_data["email"]
    
    dob = convert_timestamp(user_data["dob"]["date"])

    username = user_data["email"].replace("@example.com" , "").replace(".", "") + str(dob['year'])
    
    email = TempMail.generate_email(username)

    address = str(user_data["location"]["street"]["number"]) + ", " + user_data["location"]["street"]["name"] + ", " + user_data["location"]["city"] + ", " +user_data["location"]["state"] +  ", " + user_data["location"]["country"] 

    
    return {
     "name": user_data["name"]["first"] + " " + user_data["name"]["last"], 
     "first_name": user_data["name"]["first"],
     "last_name": user_data["name"]["last"],
     "username": username, 
     "email": email,
     "password":process_password(user_data["login"]["password"]),
     "uuid": user_data["login"]["uuid"], 
     "profile_picture" : user_data["picture"]["large"], 
     "phone" : user_data["phone"],
     "address" : address,
     "gender":user_data["gender"],    
     "country": user_data["nat"],
     "dob" : merge_dicts_in_one_dict(dob, user_data["dob"]) , 
    #  "created_at" : datetime_to_str(datetime.now())
    }



def generate_persons(count: int, gender: Gender = Gender.BOTH, country: Country = None):

    if not isinstance(count, int) or count < 1:
        count = 1

    params = {'gender': gender, 'results': count}
    if country is not None:
         params['nat'] = country
    
    r = requests.get('https://randomuser.me/api', params=params)
    if r.status_code != 200:
        raise ProblemWithGetPersonData()

    results = r.json()['results']

    final = [__get_person(data) for data in results]

    ids =  [{"id":  1 + i } for i in range(len(results))]
    
    datas = merge_list_of_dicts(  ids, final)
    
    return datas



class AccountGenerator:
    def generate_account(gender: Gender = Gender.BOTH, country: Country = None):
            return generate_persons(1, gender, country)[0]

    def generate_accounts(n, gender: Gender = Gender.BOTH, country: Country = None):
            return generate_persons(n, gender, country)



import os
import json

from .utils import relative_path


class localStoragePyStorageException(Exception):
    pass


class BasicStorageBackend:
    def raise_dummy_exception(self):
        raise localStoragePyStorageException("Called dummy backend!")

    def get_item(self, item: str, default:any = None) -> str:
        self.raise_dummy_exception()

    def set_item(self, item: str, value: any) -> None:
        self.raise_dummy_exception()

    def remove_item(self, item: str) -> None:
        self.raise_dummy_exception()

    def clear(self) -> None:
        self.raise_dummy_exception()

class JSONStorageBackend(BasicStorageBackend):
    def __init__(self) -> None:
        self.refresh()

    def refresh(self):
        self.json_path = relative_path( "local_storage.json", 0)
        self.json_data = {}

        if not os.path.isfile(self.json_path):
            self.commit_to_disk()

        with open(self.json_path, "r") as json_file:
            self.json_data = json.load(json_file)
        
    def commit_to_disk(self):
        with open(self.json_path, "w") as json_file:
            json.dump(self.json_data, json_file, indent=4)

    def get_item(self, key: str, default = None) -> str:
        if key in self.json_data:
            return self.json_data[key]
        return default


    def items(self):
        return self.json_data

    def set_item(self, key: str, value: any) -> None:
        self.json_data[key] = value
        self.commit_to_disk()

    def remove_item(self, key: str) -> None: 
        if key in self.json_data:
            self.json_data.pop(key)
            self.commit_to_disk()


    # def get_new_number(self):
    #     seen = self.get_item('seen', [])
        
    #     if len(seen) == 0:
    #         max_seen = 0
    #     else:
    #         max_seen = max(seen)
        
    #     new =  max_seen + 1
    #     self.set_item('seen', seen + [new])
    #     return new

    def clear(self) -> None:
        if os.path.isfile(self.json_path):
            os.remove(self.json_path)
        self.json_data = {}
        self.commit_to_disk()
    
class _LocalStorage:
    def __init__(self) -> None:
        self.storage_backend_instance = JSONStorageBackend()

    def refresh(self) -> None:
        self.storage_backend_instance.refresh()
    
    def get_item(self, item: str, default = None) -> any:
        return self.storage_backend_instance.get_item(item, default)

    def set_item(self, item: str, value: any) -> None:
        self.storage_backend_instance.set_item(item, value)

    def remove_item(self, item: str) -> None:
        self.storage_backend_instance.remove_item(item)

    def clear(self):
        self.storage_backend_instance.clear()

    def items(self):
        return self.storage_backend_instance.items()


    # def get_new_number(self):
    #     return self.storage_backend_instance.get_new_number()

LocalStorage = _LocalStorage()

if __name__ == "__main__":
    t = _LocalStorage()
    
    print(t.get_item("a"))
    print(t.set_item("a" ,"ss"))
    print(t.remove_item("a"))



from .bose_driver import BoseDriver
from .base_task import BaseTask
from .base_data import BaseData
from .local_storage import LocalStorage
from .opponent import Opponent
from .user_agent import UserAgent
from .window_size import WindowSize
from .create_driver import BrowserConfig
from .wait import Wait
from .output import Output
from .task_config import TaskConfig
from .schedule_utils import ScheduleUtils
from .profile import Profile



from datetime import datetime
from .ip_utils import find_ip_details
from .utils import pretty_format_time


def format_time_diff(start_time_datetime, end_time_datetime):
    time_diff = end_time_datetime - start_time_datetime
    total_seconds = int(time_diff.total_seconds())

    minutes, seconds = divmod(total_seconds, 60)

    # Format the time difference as a string
    time_str = ""
    if minutes > 0:
        time_str += f"{minutes} {'minutes' if minutes > 1 else 'minute'}"
    if seconds >= 0:
        if time_str:
            time_str += " "
        time_str += f"{seconds} {'seconds' if seconds > 1 else 'second'}"

    return time_str

def copy_and_remove_keys(input_dict):
    # Make a shallow copy of the input dictionary
    new_dict = input_dict.copy()

    # Remove the specified keys if they exist in the dictionary
    keys_to_remove = ['end_time_datetime', 'start_time_datetime']
    for key in keys_to_remove:
        new_dict.pop(key, None)

    return new_dict

class TaskInfo():
  data = {}
  
  def start(self):
    self.data["start_time_datetime"] = datetime.now()
    pass
  
  def get_data(self):
     return copy_and_remove_keys(self.data)
  
  def end(self):
    self.data["end_time_datetime"] = datetime.now()
    self.data["duration"] = format_time_diff(self.data["start_time_datetime"],self.data["end_time_datetime"])
    

    self.data["start_time"] = pretty_format_time(self.data["start_time_datetime"])
    self.data["end_time"] = pretty_format_time(self.data["end_time_datetime"])


  def set_ip(self):
    self.data["ip_details"] = find_ip_details()

  def set_task_name(self, task_name):
    self.data["task_name"] = task_name


if __name__ == "__main__":
    t = TaskInfo()
    t.start()
    t.end()
    t.end()
    t.set_ip()
    print(t.get_data())


from .local_storage import LocalStorage
import requests

class Analytics:
    def send_tracking_data(task_name):
        try:
            cookies = {}
            headers = {}
            json_data = {
                'type': 'bose_usage',
                'data': {
                    'task_name': task_name, 
                    "count": LocalStorage.get_item('count', 0)
                },
            }

            response = requests.post('https://www.omkar.cloud/backend/actions/', 
                                    #  cookies=cookies, headers=headers, 
                                     json=json_data)
        except Exception as e:
            # raise e
            pass


import chromedriver_autoinstaller
from chromedriver_autoinstaller import get_chrome_version
from chromedriver_autoinstaller.utils import get_major_version
import os
import shutil

def is_windows():
    return os.name == 'nt'

def download_driver_in_path():
    path = chromedriver_autoinstaller.install(path='build/')  

def recreate_build_dir():
    # Get the path of the current working directory
    current_dir = os.getcwd()

    # Construct the path of the build directory
    build_dir = os.path.join(current_dir, "build")

    # Delete the build directory
    shutil.rmtree(build_dir, ignore_errors=True)

    # Create the build directory again
    os.makedirs(build_dir)


def move_driver():
    major_version = get_major_version(get_chrome_version())

    executable_name = "chromedriver.exe" if is_windows() else "chromedriver"

    def move_chromedriver():
        # Define the source and destination paths
        src_path = f"build/{major_version}/{executable_name}"
        dest_path = f"build/{executable_name}"

        # Use the shutil.move() function to move the file
        shutil.move(src_path, dest_path)
    move_chromedriver()
    shutil.rmtree(f"build/{major_version}/")


def download_driver():
    
    recreate_build_dir()
    major_version = get_major_version(get_chrome_version())
    print(f'[INFO] Downloading Driver for Chrome Version {major_version} in build/ directory. This is a one-time process. Download in progress...')

    download_driver_in_path()
    move_driver()

if __name__ == '__main__':
    download_driver()


from .utils import merge_list_of_dicts, merge_dicts_in_one_dict
import random

def generate_random_number(from_num, to_num=None):
    if to_num is None:
        return from_num
    return random.randint(from_num, to_num)

class ScheduleUtils:
    def delay(tasks, from_seconds = 30, to_seconds = None):
        delays = [ {"delay":  generate_random_number(from_seconds, to_seconds) } for i in range(len(tasks))]
        
        # result = merge_list_of_dicts(tasks, delays)
        return delays

    def delay_5s(tasks):
        return ScheduleUtils.delay(tasks, 5)


    def no_delay(tasks):
        return ScheduleUtils.delay(tasks, 0)

    def delay_30s_to_60s(tasks):
        return ScheduleUtils.delay(tasks, 30, 60)


    def delay_30m_to_60m(tasks):
        return ScheduleUtils.delay(tasks, 30 * 60, 60 * 60)

    


from selenium.common.exceptions import WebDriverException
from .user_agent import UserAgentInstance, UserAgent
from .window_size import WindowSize, WindowSizeInstance
from .utils import NETWORK_ERRORS, get_current_profile_path, is_windows, read_json, relative_path, retry_if_is_error, silentremove
from selenium.webdriver.chrome.options import Options as GoogleChromeOptions
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from undetected_chromedriver import ChromeOptions
from .bose_driver import BoseDriver
from .bose_undetected_driver import BoseUndetectedDriver
import shutil
import os


class RetryException(Exception):
    pass

class BrowserConfig:
    def __init__( self, 
                  headless=False,  
                  proxy=None,  
                  use_undetected_driver=False, 
                  block_images_fonts_css = False, 
                  profile=None, 
                  is_tiny_profile=False,
                  user_agent=None,
                  window_size=WindowSize.window_size_1920_1080, 
                  is_eager=False, 
                  ):
        self.user_agent = user_agent
        self.proxy = proxy
        self.headless = headless
        self.window_size = window_size
        self.block_images_fonts_css = block_images_fonts_css


        if profile is not None:
            self.profile = str(profile)
        else:
            self.profile = None
        self.is_eager = is_eager
        self.use_undetected_driver = use_undetected_driver

        self.is_tiny_profile = is_tiny_profile

        if self.is_tiny_profile and self.profile is None:
            raise Exception('Profile must be given when using tiny profile')


def delete_cache(driver):
    print('Deleting Cache')
    driver.command_executor._commands['SEND_COMMAND'] = (
        'POST', '/session/$sessionId/chromium/send_command'
    )
    driver.execute('SEND_COMMAND', dict(
        cmd='Network.clearBrowserCache', params={}))


def add_useragent(options, user_agent):
    options.add_argument(f'--user-agent={user_agent}')


def create_profile_path(user_id):
    PROFILES_PATH = 'profiles'
    PATH = f'{PROFILES_PATH}/{user_id}'
    path = relative_path(PATH, 0)
    return path


def delete_corrupted_files(user_id):
    is_success = silentremove(
        f'{create_profile_path(user_id)}/SingletonCookie')
    silentremove(f'{create_profile_path(user_id)}/SingletonSocket')
    silentremove(f'{create_profile_path(user_id)}/SingletonLock')

    if is_success:
        print('Fixed Profile by deleting Corrupted Files')
    else:
        print('No Corrupted Profiles Found')


def delete_profile_path(user_id):
    path = create_profile_path(user_id)
    shutil.rmtree(path, ignore_errors=True)


def add_essential_options(options, profile, window_size, user_agent):
    options.add_argument("--start-maximized")
    if window_size != WindowSize.REAL:
        if window_size == None:
            if profile == None:
                window_size = WindowSizeInstance.get_random()
            else:
                window_size = WindowSizeInstance.get_hashed(profile)
        else:
            if window_size == WindowSize.RANDOM:
                window_size = WindowSizeInstance.get_random()
            elif window_size == WindowSize.HASHED:
                window_size = WindowSizeInstance.get_hashed(profile)
            else:
                window_size = window_size

        window_size = WindowSize.window_size_to_string(window_size)
        options.add_argument(f"--window-size={window_size}")

    if profile is not None:
        profile = str(profile)

    if user_agent != UserAgent.REAL:
        if user_agent == None:
            if profile == None:
                user_agent = UserAgentInstance.get_random()
            else:
                user_agent = UserAgentInstance.get_hashed(profile)
        else:
            if user_agent == UserAgent.RANDOM:
                user_agent = UserAgentInstance.get_random()
            elif user_agent == UserAgent.HASHED:
                user_agent = UserAgentInstance.get_hashed(profile)
            else:
                user_agent = user_agent

        add_useragent(options, user_agent)

    has_user = profile is not None

    if has_user:
        path = create_profile_path(profile)
        user_data_path = f"--user-data-dir={path}"
        options.add_argument(user_data_path)

    return {"window_size": window_size, "user_agent": user_agent, "profile": profile}


def get_eager_startegy():

    caps = DesiredCapabilities().CHROME
    # caps["pageLoadStrategy"] = "normal"  #  Waits for full page load
    caps["pageLoadStrategy"] = "none"   # Do not wait for full page load
    return caps


def hide_automation_flags(options):
    options.add_argument('--disable-blink-features=AutomationControlled')
    options.add_argument("--disable-blink-features")

    options.add_experimental_option(
        "excludeSwitches", ["enable-automation"])
    options.add_experimental_option('useAutomationExtension', False)

    # New Options
    options.add_argument("--ignore-certificate-errors")
    options.add_argument('--no-sandbox')
    options.add_argument("--disable-extensions")


def is_docker():
    path = '/proc/self/cgroup'

    return (
        os.path.exists('/.dockerenv') or
        os.path.isfile(path) and any('docker' in line for line in open(path))
        or os.environ.get('KUBERNETES_SERVICE_HOST') is not None
    )


def get_driver_path():
    executable_name = "chromedriver.exe" if is_windows() else "chromedriver"
    dest_path = f"build/{executable_name}"
    return dest_path


def load_cookies(driver: BoseDriver, config):
    current_profile = get_current_profile_path(config)
    current_profile_path = relative_path(current_profile, 0)

    if not os.path.exists(current_profile_path):
        os.makedirs(current_profile_path)

    current_profile_data = get_current_profile_path(config) + 'profile.json'
    current_profile_data_path = relative_path(current_profile_data, 0)

    if not os.path.isfile(current_profile_data_path):
        return

    cookies = read_json(current_profile_data_path)
    # Enables network tracking so we may use Network.setCookie method
    driver.execute_cdp_cmd('Network.enable', {})
    # Iterate through pickle dict and add all the cookies
    for cookie in cookies:
        # Fix issue Chrome exports 'expiry' key but expects 'expire' on import
        if 'expiry' in cookie:
            cookie['expires'] = cookie['expiry']
            del cookie['expiry']
        # Replace domain 'apple.com' with 'microsoft.com' cookies
        cookie['domain'] = cookie['domain'].replace(
            'apple.com', 'microsoft.com')
        # Set the actual cookie
        driver.execute_cdp_cmd('Network.setCookie', cookie)

    driver.execute_cdp_cmd('Network.disable', {})


def create_driver(config: BrowserConfig):
    def run():
        is_undetected = config.use_undetected_driver
        options = ChromeOptions() if is_undetected else GoogleChromeOptions()

        if config.headless:
            options.add_argument('--headless=new')

        selwireOptions = None
        if config.proxy is not None:
                selwireOptions = {'proxy': {'http': config.proxy, 'https': config.proxy}}
        
        if is_docker():
            print("Running in Docker, So adding sandbox arguments")
            options.arguments.extend(
                ["--no-sandbox", "--disable-setuid-sandbox"])

        if config.block_images_fonts_css:
            options.add_experimental_option(
                "prefs", {
                    "profile.managed_default_content_settings.images": 2,
                    "profile.managed_default_content_settings.stylesheet": 2,
                    "profile.managed_default_content_settings.fonts": 2,
                }
            )

        driver_attributes = add_essential_options(
            options, None if config.is_tiny_profile else config.profile, config.window_size, config.user_agent)

        if driver_attributes["profile"] is not None:
            driver_string = "Creating Driver with profile {}, window_size={}, and user_agent={}".format(
                driver_attributes["profile"], driver_attributes["window_size"], driver_attributes["user_agent"])
        else:
            driver_string = "Creating Driver with window_size={} and user_agent={}".format(
                driver_attributes["window_size"], driver_attributes["user_agent"])

        if config.is_eager:
            desired_capabilities = get_eager_startegy()
        else:
            desired_capabilities = None

        print(driver_string)

        if is_undetected:
            if selwireOptions is not None:
                raise Exception("Cannot use proxy with Undetected Driver")
            driver = BoseUndetectedDriver(
                desired_capabilities=desired_capabilities,
                options=options
            )
        else:
            # options.add_experimental_option("prefs",  {"profile.managed_default_content_settings.images": 2})
            hide_automation_flags(options)

            # CAPTCHA
            options.arguments.extend(
                ["--disable-web-security", "--disable-site-isolation-trials", "--disable-application-cache"])

            path = relative_path(get_driver_path(), 0)
            if selwireOptions is not None:
                from .bose_driver_selenium_wire import BoseDriverSeleniumWire

                driver = BoseDriverSeleniumWire(
                                    desired_capabilities=desired_capabilities,
                                    seleniumwire_options=selwireOptions,
                                    chrome_options=options,
                                    executable_path=path,
                                )  
            else:
                driver = BoseDriver(

                    desired_capabilities=desired_capabilities,
                    chrome_options=options,
                    executable_path=path,
                )
        if driver_attributes["profile"] is None:
            del driver_attributes["profile"]

        driver._init_data = driver_attributes
        return driver
    driver = retry_if_is_error(
        run, NETWORK_ERRORS + [(WebDriverException, lambda: delete_corrupted_files(config.profile) if config.profile else None)], 5)
    print("Launched Browser")

    if config.is_tiny_profile:
        load_cookies(driver, config)

    driver.browser_config = config
    return driver



from .local_storage import LocalStorage
from .output import Output
from .utils import sleep_for_n_seconds
from .ip_utils import get_valid_ip
from .beep_utils import beep_input

def prompt_change_ip(should_beep):
    current_ip = get_valid_ip()
    seen_ips =  LocalStorage.get_item("seen_ips", [])
    
    next_prompt = "Please change your IP and press Enter to continue..."
    
    while True:
        beep_input(next_prompt, should_beep)
        new_ip = get_valid_ip()

        if new_ip == current_ip:
            next_prompt = """In order to proceed, it is necessary to change your IP address as a precautionary measure against Bot Detection. Please visit https://www.omkar.cloud/bose/docs/guides/change-ip/ to learn how to change your IP. Once you have successfully changed your IP address, please press Enter to continue..."""

        elif new_ip in seen_ips:
            next_prompt = "Your computer previously had this IP address. Please change your IP and press Enter to continue..."
        else:
            LocalStorage.set_item("seen_ips", LocalStorage.get_item("seen_ips", []) + [current_ip])
            return new_ip

def launch_tasks(*tasks):
    for Task in tasks: 
        print('Task Started')

        task  = Task()
        task_config = task.get_task_config()
        data = task.get_data()
        if type(data) is not list:
            data = [data]

        should_first_beep = len(tasks) > 1

        if len(data) > 0:
            if task_config.prompt_to_close_browser: 
                prompt_message = f"Kindly close other browsers where {task_config.target_website} is open to prevent detection from {task_config.target_website} and press enter to continue..."
                beep_input(prompt_message, task_config.beep and should_first_beep)

            if task_config.change_ip: 
                prompt_change_ip(task_config.beep and should_first_beep)
        
        
        schedules = task.schedule(data)

        output = []
        for index in range(len(data)):
            current_data = data[index]
            delay = schedules[index]['delay']

            current_output = task.begin_task(current_data, task_config)
            
            if type(current_output) is dict:
                current_output = [current_output]

            if type(current_output) is list:
                output = output + current_output

            if len(output) > 0: 
                Output.write_json(output, task_config.output_filename)

            if task_config.change_ip: 
                prompt_change_ip(task_config.beep)


            if delay > 0:
                is_last = index == len(data) -1
                if is_last:
                    pass
                else:
                    sleep_for_n_seconds(delay)

        print('Task Completed!')

        if len(output) > 0: 
            Output.write_json(output, task_config.output_filename)
            Output.write_csv(output, task_config.output_filename)




class TaskConfig:
    def __init__(self, target_website = 'target website' , log_time = True, prompt_to_close_browser=False, change_ip=False, beep=True, close_on_crash = False, output_filename='finished'):
        self.target_website = target_website
        self.log_time = log_time 
        self.prompt_to_close_browser = prompt_to_close_browser
        self.change_ip = change_ip
        self.beep = beep
        self.close_on_crash = close_on_crash
        self.output_filename = output_filename


from .base_data import BaseData
from .utils import is_windows, is_mac, write_json

def get_correct_agent(windows, mac, linux):
    if is_windows():
        return windows
    elif is_mac():
        return mac
    else:
        return linux

class UserAgent(BaseData):

    REAL = "REAL"
    RANDOM = "RANDOM"
    HASHED = "HASHED"
    
    google_bot = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)" 

    user_agent_106 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.37",
                                       "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36",
                                       "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.37")

    user_agent_105 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36",
                                       "Mozilla/5.0 (Macintosh; Intel Mac OS X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36",
                                       "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36")

    user_agent_104 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.102 Safari/537.36",
                                       "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
                                       "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.102 Safari/537.36")

    user_agent_103 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.53 Safari/537.36",
                                       "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.53 Safari/537.36",
                                       "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.53 Safari/537.36")

    user_agent_101 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951 Safari/537.36",
                                       "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.0 Safari/537.36",
                                       "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951 Safari/537.36"
                                       )

    user_agent_100 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896 Safari/537.36",
                                       "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896.62 Safari/537.36",
                                       "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896 Safari/537.36"
                                       )

    user_agent_99 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844 Safari/537.36",
                                      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.51 Safari/537.36",
                                      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844 Safari/537.36"
                                      )

    user_agent_98 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758 Safari/537.36",
                                      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.80 Safari/537.36",
                                      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758 Safari/537.36"
                                      )

    user_agent_97 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692 Safari/537.36",
                                      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.71 Safari/537.36",
                                      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692 Safari/537.36"
                                      )

    user_agent_96 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664 Safari/537.36",
                                      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36",
                                      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664 Safari/537.36"
                                      )

    user_agent_95 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638 Safari/537.36",
                                      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.54 Safari/537.36",
                                      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638 Safari/537.36"
                                      )

    user_agent_94 = get_correct_agent("Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606 Safari/537.36",
                                      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.81 Safari/537.36",
                                      "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606 Safari/537.36")

    def get_data(self):
        # Versions
        V_106 = 37
        V_105 = 42
        V_104 = 2
        V_103 = 2
        V_101 = 1
        V_99 = 10
        V_100 = 1
        V_98 = 1
        V_97 = 1
        V_96 = 1
        V_95 = 1
        V_94 = 1

        _106 = [self.user_agent_106] * V_106
        _105 = [
            self.user_agent_105] * V_105
        _104 = [
            self.user_agent_104] * V_104
        _103 = [
            self.user_agent_103] * V_103
        _101 = [
            self.user_agent_101] * V_101
        _99 = [
            self.user_agent_99] * V_99
        _100 = [
            self.user_agent_100] * V_100
        _98 = [
            self.user_agent_98] * V_98
        _97 = [
            self.user_agent_97] * V_97
        _96 = [
            self.user_agent_96] * V_96
        _95 = [
            self.user_agent_95] * V_95
        _94 = [self.user_agent_94] * V_94

        result = _106 + _105 + _104 + _103 + \
            _101 + _99 + _100 + _98 + _97 + _96 + _95 + _94

        return result


UserAgentInstance = UserAgent()
if __name__ == '__main__':
    UserAgentInstance = UserAgent()
    print(UserAgentInstance.get_random_cycled())
    print(UserAgentInstance.get_random_cycled())
    print(UserAgentInstance.get_random_cycled())



from datetime import datetime
import errno
import json
import sys
from time import sleep
import traceback
from urllib.error import ContentTooShortError
from http.client import RemoteDisconnected
from urllib.error import ContentTooShortError, URLError
from sys import platform
import os

def is_mac():
    return platform == "darwin"

def is_linux():
    return platform == "linux" or platform == "linux2"

def is_windows():
    return os.name == 'nt'

def relative_path(path, goback=0):
    levels = [".."] * (goback + -1)
    return os.path.abspath(os.path.join(os.getcwd(), *levels, path))



def retry(func, retry_wait=None, retries=5):
    tries = 0
    while tries < retries:
        tries += 1
        try:
            created_result = func()
            return created_result
        except Exception as e:

            traceback.print_exc()

            if tries == retries:
                raise e

            print('Retrying')

            if retry_wait is not None:
                sleep(retry_wait)


def is_errors_instance(instances, error):
    for i in range(len(instances)):
        ins = instances[i]
        if isinstance(error, ins):
            return True, i
    return False, -1


def istuple(el):
    return type(el) is tuple


def ignore_errors(func, instances=None):
    try:
        created_result = func()
        return created_result
    except Exception as e:
        is_valid_error, index = is_errors_instance(
            instances, e)
        if not is_valid_error:
            raise e
        print('Ignoring')
        traceback.print_exc()


def retry_if_is_error(func, instances=None, retries=2, wait_time=None):
    tries = 0
    errors_only_instances = list(
        map(lambda el: el[0] if istuple(el) else el, instances))

    while tries < retries:
        tries += 1
        try:
            created_result = func()
            return created_result
        except Exception as e:
            is_valid_error, index = is_errors_instance(
                errors_only_instances, e)

            if not is_valid_error:
                raise e

            traceback.print_exc()

            if istuple(instances[index]):
                instances[index][1]()

            if tries == retries:
                raise e

            print('Retrying')

            if wait_time is not None:
                sleep(wait_time)


def keep_doing(func, wait=1):
    while True:
        func()
        sleep(wait)


def silentremove(filename):
    try:
        os.remove(filename)
        return True
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        else:
            return False

def get_range(_from, _to):
    return ",".join(map(str, list(range(_from, _to))))


def find_by_key(ls,  key, value):
    for index in range(len(ls)):
        user = ls[index]
        if user[key] == value:
            return user


def find_by_id(ls, id):
    return find_by_key(ls, 'id', id)


def remove_nones(list):
    return [element for element in list if element is not None]





def exit_with_failed_status():
    print('Exiting with status 1')
    sys.exit(1)




def sleep_for_n_seconds(n):
    print(f"Sleeping for {n} seconds...")
    sleep(n)


def on_exception(f, on_exception):
    try:
        f()
    except Exception as e:
        print(e)
        on_exception()


def merge_list_of_dicts(*dicts):
    result = []
    for i in range(len(dicts[0])):
        el = {}
        for each_dict in dicts:
            el.update(each_dict[i])
        result.append(el)
    return result


def merge_dicts_in_one_dict(*dicts):
    el = {}
    for each_dict in dicts:
        el.update(each_dict)
    return el


def wrap_with_dict(ls, key):
    def wrap(i):
        return {f'{key}': i}
    return list(map(wrap, ls))


def delete_from_dicts(ls, key):
    for x in ls:
        x.pop(key)
    return ls


def extract_from_dict(ls, key):
    return list(map(lambda i: i[key], ls))



def get_boolean_variable(name: str, default_value: bool = None):
    # Add more entries if you want, like: `y`, `yes`, ...
    true_ = ('True', 'true', '1', 't')
    false_ = ('False', 'false', '0', 'f')
    value = os.getenv(name, None)
    if value is None:
        if default_value is None:
            raise ValueError(f'Variable `{name}` not set!')
        else:
            value = str(default_value)
    if value.lower() not in true_ + false_:
        raise ValueError(f'Invalid value `{value}` for variable `{name}`')
    return value in true_


def pretty_print(result):
    print(json.dumps(result, indent=4))


def sleep_forever():
    print('Sleeping Forever')
    while True:
        sleep(100)


def flatten(l):
    return [item for sublist in l for item in sublist]


def is_error(errs):
    def fn(e):
        result, index = is_errors_instance(errs, e)
        return result
    return fn

NETWORK_ERRORS = [RemoteDisconnected, URLError,
                  ConnectionAbortedError, ContentTooShortError,  BlockingIOError]

is_network_error = is_error(NETWORK_ERRORS)

def pretty_format_time(time):
   return time.strftime("%H:%M:%S, %d %B %Y").replace(" 0", " ").lstrip("0")



def write_html( data, path,):
    with open(path, 'w', encoding="utf-8") as fp:
        fp.write(data)


def write_file( data, path,):
    with open(path, 'w', encoding="utf-8") as fp:
        fp.write(data)


def read_json(path):
    with open(path, 'r') as fp:
        users = json.load(fp)
        return users

def read_file(path):
    with open(path, 'r') as fp:
        content = fp.read()
        return content
        
def write_json(data, path,  indent=4):
    with open(path, 'w') as fp:
        json.dump(data, fp, indent=indent)


def get_driver_path():
    executable_name = "chromedriver.exe" if is_windows() else "chromedriver"
    dest_path = f"build/{executable_name}"
    return dest_path


datetime_format = '%Y-%m-%d %H:%M:%S'

def str_to_datetime(when):
    return datetime.strptime(
        when, datetime_format)


def datetime_to_str(when):
    return when.strftime(datetime_format)


def get_current_profile_path(config): 
    profiles_path = f'profiles/{config.profile}/'
    # profiles_path =  relative_path(path, 0)
    return profiles_path



import requests
from requests.exceptions import ReadTimeout
import traceback
def find_ip_details(max_retries=5):
    url = 'https://www.omkar.cloud/backend/ipinfo/'
    try:
        response = requests.get(url, timeout=10)
        data = response.json()

        if "readme" in data:
           del data["readme"]
        return data
    except requests.exceptions.ReadTimeout:
        if max_retries > 0:
            print('ReadTimeout occurred. Retrying...')
            return find_ip_details(max_retries - 1)
        else:
            print('Max retries exceeded. Unable to fetch IP details.')
            return None
    except Exception as e:
        traceback.print_exc()
        return None

def find_ip(attempts=5):
    url = 'https://checkip.amazonaws.com/'
    try:
        response = requests.get(url, timeout=10)
        return response.text.strip()

    except ReadTimeout:
        if attempts > 1:
            print("ReadTimeout occurred. Retrying...")
            return find_ip(attempts - 1)
        else:
            print("Max attempts reached. Failed to get IP address.")
            return None

    except Exception as e:
        traceback.print_exc()
        return None


def get_valid_ip():
    ip = find_ip()
    while ip is None:
        print("Failed to get IP. Retrying...")
        ip = find_ip()
    return ip



class LocalStorage:
    def __init__(self, driver):
        self.driver = driver

    def __len__(self):
        return self.driver.execute_script("return window.localStorage.length;")

    def items(self):
        return self.driver.execute_script(
            "var ls = window.localStorage, items = {}; "
            "for (var i = 0, k; i < ls.length; ++i) "
            "  items[k = ls.key(i)] = ls.getItem(k); "
            "return items; ")

    def keys(self):
        return self.driver.execute_script(
            "var ls = window.localStorage, keys = []; "
            "for (var i = 0; i < ls.length; ++i) "
            "  keys[i] = ls.key(i); "
            "return keys; ")

    def get_item(self, key):
        return self.driver.execute_script("return window.localStorage.getItem(arguments[0]);", key)

    def set_item(self, key, value):
        self.driver.execute_script(
            "window.localStorage.setItem(arguments[0], arguments[1]);", key, value)

    def has_item(self, key):
        return key in self.keys()

    def remove_item(self, key):
        self.driver.execute_script(
            "window.localStorage.removeItem(arguments[0]);", key)

    def clear(self):
        self.driver.execute_script("window.localStorage.clear();")

    def __getitem__(self, key):
        value = self.get_item(key)
        if value is None:
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):
        self.set_item(key, value)

    def __contains__(self, key):
        return key in self.keys()

    def __iter__(self):
        return self.items().__iter__()

    def __repr__(self):
        return self.items().__str__()


from distutils.core import setup


install_requires = [
    "requests",
    "chromedriver-autoinstaller==0.4.0",
    "selenium==4.5.0",
    "undetected_chromedriver>=3.4.6",
    "beautifulsoup4>=4.11.2",
    "joblib==1.2.0",
    "selenium-wire",
]
extras_require = {}
cpython_dependencies = [
    "PyDispatcher>=2.0.5",
]

def get_description():
    try:
      return open("README.rst", encoding="utf-8").read()
    except:
      return None
    
            
setup(
    name='bose',
    packages=['bose'],
    version='2.0.11',
    license='MIT',
    project_urls={
        "Documentation": "https://omkar.cloud/bose/",
        "Source": "https://github.com/omkarcloud/bose",
        "Tracker": "https://github.com/omkarcloud/bose/issues",
    },

    description="The Ultimate Web Scraping Framework",
    long_description=get_description(),
    # long_description_content_type="text/markdown",
    author='Chetan Jain',
    author_email='chetan@omkar.cloud',
    maintainer="Chetan Jain",
    maintainer_email="chetan@omkar.cloud",

    keywords=["crawler", "framework", "scraping", "crawling", "web-scraping", "web-scraping-python", "cloudflare-bypass", "anti-detection", "bot-detection", "automation", "webdriver", "browser"],
    classifiers=[
        "Framework :: Scrapy",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.7",
    install_requires=install_requires,
    extras_require=extras_require,
)


import traceback
from datetime import datetime
from seleniumwire import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
import random
from time import sleep
from .beep_utils import beep_input
from .local_storage_driver import LocalStorage
from .opponent import Opponent
from .utils import get_current_profile_path, read_file, relative_path, sleep_for_n_seconds, sleep_forever, write_json

    
def save_cookies(driver, config):
            current_profile_data = get_current_profile_path(config) + 'profile.json'
            current_profile_data_path =  relative_path(current_profile_data, 0)

            driver.execute_cdp_cmd('Network.enable', {})
            cookies = (driver.execute_cdp_cmd('Network.getAllCookies', {}))
            driver.execute_cdp_cmd('Network.disable', {})

            if type(cookies) is not list:
                cookies = cookies.get('cookies')
            write_json(cookies, current_profile_data_path)


class BoseDriverSeleniumWire(webdriver.Chrome):
    beep = True

    def get_by_current_page_referrer(self, link, wait=None):

        # selenium.common.exceptions.WebDriverException
        self.execute_script(f"""
                window.location.href = "{link}";
            """)
        if wait is not None and wait != 0:
            sleep(wait)

    def js_click(self, element):
        self.execute_script("arguments[0].click();",  element)

    def sleep(self, n):
        sleep_for_n_seconds(n)

    def prompt(self, text="Press Enter To Continue..."):
        return beep_input(text, self.beep)

    def short_random_sleep(self):
        sleep_for_n_seconds(random.uniform(2, 4))

    def long_random_sleep(self):
        sleep_for_n_seconds(random.uniform(6, 9))

    def sleep_forever(self):
        sleep_forever()

    def get_bot_detected_by(self):

        pmx = self.get_element_or_none(
            "//*[text()='Please verify you are a human']")
        if pmx is not None:
            return Opponent.PERIMETER_X

        clf = self.get_element_or_none_by_selector("#challenge-running")
        if clf is not None:
            return Opponent.CLOUDFLARE

        return None

    def is_bot_detected(self):
        return self.get_bot_detected_by() is not None

    def get_element_or_none(self, xpath, wait=None) -> WebElement:
        try:
            if wait is None:
                return self.find_element(By.XPATH, xpath)
            else:
                return WebDriverWait(self, wait).until(
                    EC.presence_of_element_located((By.XPATH, xpath)))
        except:
            return None

    def get_element_or_none_by_selector(self: WebDriver, selector, wait=None) -> WebElement:
        try:
            if wait is None:
                return self.find_element(By.CSS_SELECTOR, selector)
            else:
                return WebDriverWait(self, wait).until(
                    EC.element_to_be_clickable((By.CSS_SELECTOR, selector)))
        except:
            return None

    def get_element_by_id(self, id: str, wait=None):
        cleaned = id.lstrip('#')
        return self.get_element_or_none_by_selector(f'[id="{cleaned}"]', wait)

    def get_element_or_none_by_text_contains(self, text, wait=None):
        text = f'//*[contains(text(), "{text}")]'
        return self.get_element_or_none(text, wait)

    def get_element_or_none_by_text(self, text, wait=None):
        text = f'//*[text()="{text}"]'

        return self.get_element_or_none(text, wait)

    def get_element_parent(element):
        return element.find_element(By.XPATH, "./..")

    def get_elements_or_none_by_selector(self: WebDriver, selector, wait=None):
        try:
            if wait is None:
                return self.find_elements(By.CSS_SELECTOR, selector)
            else:
                WebDriverWait(self, wait).until(
                    EC.presence_of_element_located((By.CSS_SELECTOR, selector)))

                return self.find_elements(By.CSS_SELECTOR, selector)
        except:
            return None

    def get_element_text(self, element):
        return element.get_attribute('innerText')

    def get_innerhtml(self, element):
        return element.get_attribute("innerHTML")

    def get_element_or_none_by_name(self, selector, wait=None):
        try:
            if wait is None:
                return self.find_element(By.NAME, selector)
            else:
                return WebDriverWait(self, wait).until(
                    EC.presence_of_element_located((By.NAME, selector)))
        except:
            return None

    def scroll_site(self):
        self.execute_script(""" 
window.scrollBy(0, 10000);
""")

    def can_element_be_scrolled(self, element):
        # <=3 is a fix to handle floating point numbers
        result = not (self.execute_script(
            "return Math.abs(arguments[0].scrollTop - (arguments[0].scrollHeight - arguments[0].offsetHeight)) <= 3", element))
        return result

    def scroll_element(self, element):
        if self.can_element_be_scrolled(element):
            self.execute_script("arguments[0].scrollBy(0, 10000)", element)
            return True
        else:
            return False

    def get_cookies_dict(self):
        all_cookies = self.get_cookies()
        cookies_dict = {}
        for cookie in all_cookies:
            cookies_dict[cookie['name']] = cookie['value']
        return cookies_dict

    def get_local_storage_dict(self):
        storage = LocalStorage(self)
        return storage.items()

    def get_cookies_and_local_storage_dict(self):
        cookies = self.get_cookies_dict()
        local_storage = self.get_local_storage_dict()

        return {"cookies": cookies, "local_storage": local_storage}

    def add_cookies_dict(self, cookies):
        for key in cookies:
            self.add_cookie({"name": key, "value": cookies[key]})

    def add_local_storage_dict(self, local_storage):
        storage = LocalStorage(self)
        for key in local_storage:
            storage.set_item(key, local_storage[key])

    def add_cookies_and_local_storage_dict(self, site_data):
        cookies = site_data["cookies"]
        local_storage = site_data["local_storage"]
        self.add_cookies_dict(cookies)
        self.add_local_storage_dict(local_storage)

    def delete_cookies_dict(self):
        self.delete_all_cookies()

    def delete_local_storage_dict(self):
        self.execute_script("window.localStorage.clear();")
        self.execute_script("window.sessionStorage.clear();")

    def delete_cookies_and_local_storage_dict(self):
        self.delete_cookies_dict()
        self.delete_local_storage_dict()

    def organic_get(self, link, wait=None):
        self.get("https://www.google.com/")
        self.get_by_current_page_referrer(link, wait)

    def get_google(self):
        self.get("https://www.google.com/")
        # self.get_element_or_none_by_selector('input[role="combobox"]', Wait.VERY_LONG)

    @property
    def local_storage(self):
        return LocalStorage(self)

    def get_links(self, search=None, wait=None):

        def extract_links(elements):
            def extract_link(el):
                return el.get_attribute("href")

            return list(map(extract_link, elements))

        els = self.get_elements_or_none_by_selector("a", wait)

        links = extract_links(els)

        def is_not_none(link):
            return link is not None

        def is_starts_with(link):
            if search == None:
                return True
            return search in link

        return list(filter(is_starts_with, filter(is_not_none, links)))



    def execute_file(self, filename):
        if not filename.endswith(".js"):
            filename = filename + ".js"
        content = read_file(filename)
        return self.execute_script(content)

    def get_images(self, search=None, wait=None):

        def extract_links(elements):
            def extract_link(el):
                return el.get_attribute("src")

            return list(map(extract_link, elements))

        els = self.get_elements_or_none_by_selector("img", wait)

        links = extract_links(els)

        def is_not_none(link):
            return link is not None

        def is_starts_with(link):
            if search == None:
                return True
            return search in link

        return list(filter(is_starts_with, filter(is_not_none, links)))

    def is_in_page(self, target, wait=None, raise_exception=False):

        def check_page(driver, target):
            if isinstance(target, str):
                return target in driver.current_url
            else:
                for x in target:
                    if x in driver.current_url:
                        return True
                return False

        if wait is None:
            return check_page(self, target)
        else:
            time = 0
            while time < wait:
                if check_page(self, target):
                    return True

                sleep_time = 0.2
                time += sleep_time
                sleep(sleep_time)

        if raise_exception:
            raise Exception(f"Page {target} not found")
        return False

    def save_screenshot(self, filename=datetime.now().strftime('%Y-%m-%d_%H-%M-%S') + ".png"):
        try:

            if not filename.endswith(".png"):
                filename = filename + ".png"

            final_path = f'{self.task_path}/{filename}'
            saving_screenshot_at = relative_path(
                final_path, 0)
            self.get_screenshot_as_file(
                saving_screenshot_at)
            # print('Saved screenshot at {0}'.format(final_path))
        except:
            traceback.print_exc()
            print('Failed to save screenshot')

    def prompt_to_solve_captcha(self, more_rules=[]):
        print('')
        print('   __ _ _ _    _                          _       _           ')
        print('  / _(_) | |  (_)                        | |     | |          ')
        print(' | |_ _| | |   _ _ __      ___ __ _ _ __ | |_ ___| |__   __ _ ')
        print(' |  _| | | |  | | `_ \    / __/ _` | `_ \| __/ __| `_ \ / _` |')
        print(' | | | | | |  | | | | |  | (_| (_| | |_) | || (__| | | | (_| |')
        print(' |_| |_|_|_|  |_|_| |_|   \___\__,_| .__/ \__\___|_| |_|\__,_|')
        print('                                   | |                        ')
        print('                                   |_|                        ')
        print('')

        print('General Rules of Captcha Solving')
        print(' - Solve it Fast')

        for t in more_rules:
            print(t)
        # print('- Solve it Fast')
        # print('1. If')

        return beep_input('Press fill in the captcha, the faster the less detectable, then press enter to continue ...', self.beep)

        # return beep_input('Press fill in the captcha and press enter to continue ...', self.beep)
    def close(self) -> None:
        if (self.browser_config.is_tiny_profile):
            save_cookies(self, self.browser_config)

        return super().close()



import requests
import csv
from .utils import read_file, relative_path, write_json, read_json, write_html

class Output:

    def read_json(filename):
        if not filename.startswith("output/"):
            filename = "output/" +  filename

        if not filename.endswith(".json"):
            filename = filename + ".json"

        return read_json(filename)

    def write_json(data, filename):
        if len(data) == 0:
            print("Data is empty.")
            return

        if not filename.startswith("output/"):
            filename = "output/" +  filename

        if not filename.endswith(".json"):
            filename = filename + ".json"

        write_json(data, filename)
        print(f"View written JSON file at {filename}")        

    def write_csv(data, filename):
        
        """
        Save a list of dictionaries as a CSV file.

        Args:
            data: a list of dictionaries
            filename: the name of the CSV file to save
        """
        
        if type(data) is dict:
            data = [data]

        if len(data) == 0:
            print("Data is empty.")
            return

        if not filename.startswith("output/"):
            filename = "output/" +  filename

        if not filename.endswith(".csv"):
            filename = filename + ".csv"

        with open(filename, 'w', newline='', encoding='utf-8') as csvfile:
            fieldnames = data[0].keys()  # get the fieldnames from the first dictionary
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()  # write the header row
            writer.writerows(data)  # write each row of data
        print(f"View written CSV file at {filename}")     


    def read_finished_json():
        return Output.read_json("finished.json")

    def write_finished_json(data):
        return Output.write_json(data, 'finished.json')


    def write_finished_csv(data):
        return Output.write_csv(data, 'finished.csv')

    def save_image(url, filename = None):
        if filename is None:
            filename = url.split("/")[-1]
    
        response = requests.get(url)
        if response.status_code == 200:
            # Extract the filename from the URL
            output_dir = "output"

            # Save the image in the output directory
            path = f'{output_dir}/{filename}'
            with open(relative_path(
                path, 0), "wb") as f:
                f.write(response.content)
        else:
            print("Failed to download the image.")
            return path
        
    def write_html(data, filename):
        if not filename.startswith("output/"):
            filename = "output/" +  filename

        if not filename.endswith(".html"):
            filename = filename + ".html"

        write_html(data, filename)
        print(f"View written HTML file at {filename}")


    def read_html(data, filename):
        if not filename.startswith("output/"):
            filename = "output/" +  filename

        if not filename.endswith(".html"):
            filename = filename + ".html"

        return read_file(filename)


    def write_file(data, filename):
        if not filename.startswith("output/"):
            filename = "output/" +  filename


        write_html(data, filename)
        print(f"View written file at {filename}")


    def read_file(data, filename):
        if not filename.startswith("output/"):
            filename = "output/" +  filename

        return read_file(filename)




import random as random_module
from datetime import datetime
import os
import json
from .utils import datetime_to_str, relative_path, str_to_datetime

class ProfilePyStorageException(Exception):
    pass

class BasicStorageBackend:
    def raise_dummy_exception(self):
        raise ProfilePyStorageException("Called dummy backend!")

    def get_item(self, item: str, default:any = None) -> str:
        self.raise_dummy_exception()

    def set_item(self, item: str, value: any) -> None:
        self.raise_dummy_exception()

    def remove_item(self, item: str) -> None:
        self.raise_dummy_exception()

    def clear(self) -> None:
        self.raise_dummy_exception()

class JSONStorageBackend(BasicStorageBackend):
    def __init__(self) -> None:
        self.refresh()

    def refresh(self):
        self.json_path = relative_path( "profiles.json", 0)
        self.json_data = {}

        if not os.path.isfile(self.json_path):
            self.commit_to_disk()

        with open(self.json_path, "r") as json_file:
            self.json_data = json.load(json_file)
        
    def commit_to_disk(self):
        with open(self.json_path, "w") as json_file:
            json.dump(self.json_data, json_file, indent=4)

    def get_item(self, key: str, default = None) -> str:
        if key in self.json_data:
            return self.json_data[key]
        return default

    def items(self):
        return self.json_data

    def set_item(self, key: str, value: any) -> None:
        if 'created_at' not in value:
            value['created_at'] = datetime_to_str(datetime.now())

        value['updated_at'] = datetime_to_str(datetime.now())

        self.json_data[key] = value
        self.commit_to_disk()

    def remove_item(self, key: str) -> None: 
        if key in self.json_data:
            self.json_data.pop(key)
            self.commit_to_disk()

    def clear(self) -> None:
        if os.path.isfile(self.json_path):
            os.remove(self.json_path)
        self.json_data = {}
        self.commit_to_disk()
    
class _Profile:
    def __init__(self) -> None:
        self.storage_backend_instance = JSONStorageBackend()

    profile = None
    def refresh(self) -> None:
        self.storage_backend_instance.refresh()

    def check_profile(self) -> None:
        if self.profile is None:
            raise Exception('This method can only be run in run method of Task and when you have given the current profile in the Browser Config.')

    def get_item(self, item: str, default = None) -> any:
        self.check_profile()
        profile = self.storage_backend_instance.get_item(self.profile, {})

        if default is None:
            return profile.get(item)
        else: 
            return profile.get(item, default)

    def set_item(self, item: str, value: any) -> None:
        self.check_profile()
        profile = self.storage_backend_instance.get_item(self.profile, {})
        profile[item] = value

        self.storage_backend_instance.set_item(self.profile, profile)

    def remove_item(self, item: str) -> None:
        self.check_profile()
        profile = self.storage_backend_instance.get_item(self.profile, {})
        del profile[item] 

        self.storage_backend_instance.set_item(self.profile, profile)

    def clear(self):
        self.check_profile()
        self.storage_backend_instance.remove_item(self.profile)

    def items(self):
        self.check_profile()
        profile = self.storage_backend_instance.get_item(self.profile, {})
        return profile

    def get_profiles(self, random = False):
        data = list(self.storage_backend_instance.items().values())
        
        if len(data) == 0:
            return data

        if (data[0].get('created_at') is None):
            if random:
                random_module.shuffle(data)
            return data
        
        if random:
            random_module.shuffle(data)
        else: 
            data =  sorted(data, key=lambda x: str_to_datetime(x['created_at']))

        return data

    def set_profile(self, profile):
        self.check_profile()
        if type(profile) is dict:
            self.storage_backend_instance.set_item(self.profile, profile)
        else: 
            raise Exception("Profile is not a dictionary.")

Profile = _Profile()

if __name__ == "__main__":
    t = _Profile()
    
    print(t.get_item("a"))
    print(t.set_item("a" ,"ss"))
    print(t.remove_item("a"))
